from __future__ import annotations

from abc import ABC, abstractmethod

# import jnp
import numpy as np
from sklearn.metrics import mean_squared_error
import jax.numpy as jnp

from JOPLEn.enums import LossType

from .enums import DTYPE


def sigmoid(x: jnp.ndarray) -> jnp.ndarray:
    return jnp.reciprocal(1 + jnp.exp(-x))


class Loss(ABC):
    def __init__(self: Loss, loss_type: LossType) -> None:
        self.loss_type = loss_type

    @abstractmethod
    def __call__(
        self: Loss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> float:
        pass

    @abstractmethod
    def grad(
        self: Loss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        pass

    @abstractmethod
    def predict(
        self: Loss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        pass

    def _raw_output(
        self: Loss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        return jnp.sum((x @ w) * s, axis=1, keepdims=True)

    @abstractmethod
    def error(
        self: Loss,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> float:
        pass


class SquaredError(Loss):
    def __init__(self: SquaredError) -> None:
        super().__init__(LossType.regression)

    def __call__(
        self: SquaredError,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> float:
        y_pred = self.predict(w, x, s)

        return float(jnp.mean((y_pred - y) ** 2))

    def grad(
        self: SquaredError,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        y_pred = self._raw_output(w, x, s)
        return x.T @ ((y_pred - y) * s) / x.shape[0]

    def predict(
        self: SquaredError,
        w: jnp.ndarray,
        x: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        return self._raw_output(w, x, s)

    def error(
        self: SquaredError,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> float:
        return float(mean_squared_error(y_true, y_pred, squared=False))


class LogisticLoss(Loss):
    def __init__(self: LogisticLoss) -> None:
        super().__init__(LossType.binary_classification)

    def encode(self: LogisticLoss, y: np.ndarray) -> np.ndarray:
        return (y * 2) - 1

    def __call__(
        self: LogisticLoss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> float:
        raw_output = self._raw_output(w, x, s)
        y = self.encode(y)

        return float(jnp.mean(jnp.log(1 + jnp.exp(-y * raw_output))))

    def grad(
        self: LogisticLoss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        raw_output = self._raw_output(w, x, s)
        y = self.encode(y)

        return -x.T @ ((y / (jnp.exp(raw_output * y) + 1)) * s) / x.shape[0]

    def predict(
        self: LogisticLoss,
        w: jnp.ndarray,
        x: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        return sigmoid(self._raw_output(w, x, s))

    def error(
        self: LogisticLoss,
        y_true: np.ndarray,
        y_pred: np.ndarray,
    ) -> float:
        return np.mean((y_true > 0) == (y_pred > 0))



import pytest
from JOPLEn.enums import CellModel, LossType
from JOPLEn.partitioner import (
    CBPartition,
    ExtraTreePartition,
    GBPartition,
    LGBMPartition,
    LinearForestPartition,
    RFPartition,
    VPartition,
)
from JOPLEn.singletask import JOPLEn
from JOPLEn.st_loss import LogisticLoss, SquaredError
from JOPLEn.st_penalty import (
    Group21Norm,
    GroupInf1Norm,
    L1Norm,
    NuclearNorm,
    SquaredFNorm,
)


def gen_train_data(is_classification):
    from sklearn.datasets import make_classification, make_regression
    from sklearn.model_selection import train_test_split

    kwargs = {
        "n_samples": 1000,
        "n_features": 20,
        "n_informative": 10,
        "random_state": 0,
    }

    if is_classification:
        x, y = make_classification(**kwargs, n_classes=2)
        y = y.flatten()
    else:
        x, y = make_regression(**kwargs, n_targets=1, noise=0.1)

    # Split the dataset into training and test sets
    x_train, x_val, y_train, y_val = train_test_split(
        x, y, test_size=0.2, random_state=0
    )

    return x_train, y_train, x_val, y_val


# Just makes sure that the model runs without any errors
@pytest.mark.parametrize(
    "part",
    [
        VPartition,
        ExtraTreePartition,
        LGBMPartition,
        CBPartition,
        RFPartition,
        GBPartition,
        LinearForestPartition,
    ],
)
@pytest.mark.parametrize("loss_fn", [SquaredError, LogisticLoss])
@pytest.mark.parametrize("cell_model", [CellModel.linear, CellModel.constant])
@pytest.mark.parametrize(
    "reg",
    [
        SquaredFNorm,
        Group21Norm,
        GroupInf1Norm,
        NuclearNorm,
        L1Norm,
    ],
)
def test_st_reg(part, loss_fn, cell_model, reg):
    x_train, y_train, x_val, y_val = gen_train_data(
        is_classification=(loss_fn().loss_type == LossType.binary_classification)
    )

    jp = JOPLEn(
        partitioner=part,
        n_cells=8,
        n_partitions=5,
        loss_fn=loss_fn,
        cell_model=cell_model,
        mu=1e-3,
        max_iters=100,
        early_stop=False,
        rescale=False,
        regularizers=[reg(lam=0.01)],
    )

    jp.fit(
        x_train,
        y_train,
        val_x=x_val,
        val_y=y_val,
        print_epochs=10,
    )



from typing import Any

import numpy as np
from sklearn.base import ClassifierMixin, RegressorMixin
from sklearn.ensemble._base import BaseEnsemble
from sklearn.linear_model._base import LinearClassifierMixin, LinearModel


class FriedmanRefit:
    def __init__(
        self,
        base_est_class: type[BaseEnsemble],
        refit_est_class: type[LinearModel],
    ) -> None:
        assert issubclass(
            base_est_class, BaseEnsemble
        ), "base_est_class must be an ensemble"

        is_lin_class = issubclass(refit_est_class, LinearClassifierMixin)
        is_lin_reg = issubclass(refit_est_class, RegressorMixin)
        assert (
            is_lin_class or is_lin_reg
        ), f"refit_est_class must be a linear model, given {refit_est_class}"

        assert issubclass(base_est_class, ClassifierMixin) == issubclass(
            refit_est_class, ClassifierMixin
        ), "base_est_class and refit_est_class must both be classifiers or regressors"

        self.is_classifier = issubclass(base_est_class, ClassifierMixin)

        self.base_est_class: type[BaseEnsemble] = base_est_class
        self.refit_est_class: type[LinearModel] = refit_est_class

    def pred_ind(self, X: np.ndarray) -> np.ndarray:
        # Assuming that there is only one output
        estimators = np.array(self.base_est.estimators_).reshape(-1)

        preds = np.zeros((X.shape[0], len(estimators)))

        for i, est in enumerate(estimators):
            preds[:, i] = est.predict(X)

        return preds

    def fit(
        self,
        X: np.ndarray,
        y: np.ndarray,
        base_params: dict[str, Any] | None = None,
        refit_params: dict[str, Any] | None = None,
    ):
        refit_params = refit_params or {}
        base_params = base_params or {}

        self.base_est: BaseEnsemble = self.base_est_class(**base_params)
        self.refit_est: LinearModel = self.refit_est_class(**refit_params)

        self.base_est.fit(X, y)

        preds = self.pred_ind(X)

        self.refit_est.fit(preds, y)

    def predict(self, X):
        assert hasattr(self, "base_est"), "You must fit the model first"
        assert hasattr(self, "refit_est"), "You must fit the model first"

        preds = self.pred_ind(X)

        return self.refit_est.predict(preds)

    def predict_proba(self, X):
        assert hasattr(self, "base_est"), "You must fit the model first"
        assert hasattr(self, "refit_est"), "You must fit the model first"
        assert self.is_classifier, "predict_proba is only available for classifiers"

        preds = self.pred_ind(X)

        if self.is_classifier:
            return self.refit_est.predict_proba(preds)[:, 1]
        else:
            return self.refit_est.predict_proba(preds)


if __name__ == "__main__":
    print("Testing Regression")
    from sklearn.datasets import load_diabetes
    from sklearn.ensemble import GradientBoostingRegressor
    from sklearn.linear_model import Lasso, Ridge
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import train_test_split

    def rmse(y_true: np.ndarray, y_pred: np.ndarray):
        return mean_squared_error(y_true, y_pred, squared=False)

    diabetes = load_diabetes()
    X, y = diabetes.data, diabetes.target

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    # fr = FriedmanRefit(GradientBoostingRegressor, Ridge)
    # fr.fit(X_train, y_train, refit_params={"alpha": 15.0})
    fr = FriedmanRefit(GradientBoostingRegressor, Lasso)
    fr.fit(X_train, y_train, refit_params={"alpha": 100.0})

    gb = GradientBoostingRegressor()
    gb.fit(X_train, y_train)

    norm_preds = gb.predict(X_test)
    fr_preds = fr.predict(X_test)

    print(f"GBR RMSE: {rmse(y_test, norm_preds)}")
    print(f"FR RMSE: {rmse(y_test, fr_preds)}")

    print("Testing Classification")
    from sklearn.datasets import load_breast_cancer
    from sklearn.ensemble import GradientBoostingClassifier
    from sklearn.linear_model import LogisticRegression
    from sklearn.metrics import accuracy_score

    cancer = load_breast_cancer()
    X, y = cancer.data, cancer.target

    X_train, X_test, y_train, y_test = train_test_split(
        X, y, test_size=0.2, random_state=42
    )

    fr = FriedmanRefit(GradientBoostingClassifier, LogisticRegression)
    fr.fit(X_train, y_train)

    gb = GradientBoostingClassifier()
    gb.fit(X_train, y_train)

    norm_preds = gb.predict(X_test)
    fr_preds = fr.predict(X_test)

    print(f"GBR Accuracy: {accuracy_score(y_test, norm_preds)}")
    print(f"FR Accuracy: {accuracy_score(y_test, fr_preds)}")



from __future__ import annotations

import warnings
from abc import ABC, abstractmethod
from typing import Any, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from catboost import CatBoostClassifier, CatBoostRegressor, Pool
from lightgbm import LGBMClassifier, LGBMRegressor
from lineartree import LinearForestClassifier, LinearForestRegressor
from numpy.random import RandomState
from scipy.spatial import Voronoi, voronoi_plot_2d
from sklearn.ensemble import (
    ExtraTreesClassifier,
    ExtraTreesRegressor,
    GradientBoostingClassifier,
    GradientBoostingRegressor,
    RandomForestClassifier,
    RandomForestRegressor,
)
from sklearn.linear_model import LinearRegression
from sklearn.metrics import pairwise_distances_argmin_min
from xgboost import DMatrix, XGBClassifier, XGBRegressor

from .enums import LossType


def numpify(x: np.ndarray | pd.DataFrame | pd.Series) -> np.ndarray:
    """Convert a pandas DataFrame to a numpy array if necessary.

    Args:
        x (pd.DataFrame | np.ndarray): The input data.

    Returns:
        np.ndarray: The input data as a numpy array.
    """
    if isinstance(x, pd.DataFrame) or isinstance(x, pd.Series):
        x = x.to_numpy()

    if len(x.shape) == 1:
        x = x.reshape(-1, 1)

    return x


class Partitioner(ABC):
    def __init__(
        self: Partitioner,
        n_cells: int,
        n_partitions: int,
        loss_type: LossType,
        random_state: Union[int, RandomState],
        keep_int: bool = False,
        **model_kwargs: dict[str, Any],
    ) -> None:
        # keep_int is necessary for CatBoost for some reason
        self.n_cells: int = n_cells
        self.n_partitions: int = n_partitions
        self.loss_type: LossType = loss_type
        self.model_kwargs: dict[str, Any] = model_kwargs

        if not isinstance(random_state, RandomState) and not keep_int:
            self.state: RandomState | int = RandomState(random_state)
        else:
            self.state: RandomState | int = random_state

    @abstractmethod
    def partition(self: Partitioner, x: np.ndarray) -> np.ndarray:
        """Take input data and apply a partition mask to it.

        Args:
            self (Partitioner): The partitioner object.
            x (np.ndarray): The input data to be partitioned.

        Returns:
            np.ndarray: The partitioned data.

        """


class VPartition(Partitioner):
    def __init__(
        self: VPartition,
        x: np.ndarray,
        y: np.ndarray,
        n_cells: int,
        n_partitions: int,
        loss_type: LossType,
        random_state: int = 0,
        x_val: np.ndarray | None = None,
        y_val: np.ndarray | None = None,
        **model_kwargs: dict[str, Any],
    ) -> None:
        """Initializes a VPartition object.

        Args:
            x (np.ndarray): Array of shape (n_samples, n_features) containing the input data.
            y (np.ndarray): Array of shape (n_samples,) containing the target labels.
            n_cells (int): Number of Voronoi cells to create.
            n_partitions (int): Number of partitions to create.
            random_state (int, optional): Seed for the random number generator. Defaults to 0.

        """
        super().__init__(n_cells, n_partitions, loss_type, random_state, **model_kwargs)

        self._create_voronoi(x)

    def _create_voronoi(self: VPartition, x: np.ndarray) -> None:
        idx = self.state.randint(0, x.shape[0], size=(self.n_cells, self.n_partitions))

        self.vor_points = x[idx]

    def partition(self: VPartition, x: np.ndarray) -> np.ndarray:
        """Partition the input data using Voronoi cells.

        Args:
            self (VPartition): The Voronoi partitioner object.
            x (np.ndarray): The input data to be partitioned.

        Returns:
            np.ndarray: The partitioned data wth shape (n_samples, n_partitions)
        """
        # Check if the dimensions are compatible
        if x.shape[1] != self.vor_points.shape[2]:
            msg = (
                "Dimension mismatch between input data and Voronoi points:"
                f"{x.shape[1]}, {self.vor_points.shape[2]}"
            )
            raise ValueError(msg)

        x = numpify(x)

        # Initialize an empty list to store the results
        concatenated_indices = []

        # Iterate through the n different sets of Voronoi points
        for i in range(self.vor_points.shape[1]):
            vor_points_i = self.vor_points[:, i, :]
            indices, _ = pairwise_distances_argmin_min(x, vor_points_i)
            concatenated_indices.append(indices.reshape(-1, 1))

        # Concatenate the results along the specified axis
        return np.hstack(concatenated_indices)

    def plot_partitions(self: VPartition, max_to_plot: int = 1, ax=None) -> None:
        """Plot the Voronoi partitions.

        Args:
            self (VPartition): The Voronoi partitioner object.
            max_to_plot (int, optional): The number of partitions to plot. Defaults to 1.
            ax (_type_, optional): The matplotlib axis to plot on. Defaults to None.

        """
        from matplotlib import cm

        if ax is None:
            fig, ax = plt.subplots()

        colors = cm.get_cmap("tab10").colors

        for i in range(max_to_plot):
            vor = Voronoi(self.vor_points[:, i, :])
            voronoi_plot_2d(
                vor,
                show_vertices=False,
                line_colors=colors[i],
                line_styles="-",
                point_size=0,
                ax=ax,
            )


class TreePartition(Partitioner, ABC):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        n_cells: int,
        n_partitions: int,
        loss_type: LossType,
        random_state: int = 0,
        prefit_model: Any | None = None,
        x_val: np.ndarray | None = None,
        y_val: np.ndarray | None = None,
        keep_int: bool = False,
        **model_kwargs: dict[str, Any],
    ) -> None:
        super().__init__(
            n_cells,
            n_partitions,
            loss_type,
            random_state,
            keep_int,
            **model_kwargs,
        )

        if prefit_model is None:
            self._fit_model(x, y, x_val, y_val)
        else:
            self._prefit_model(prefit_model, x)

        # should provide all of the leaf indices
        train_leaf_indices = self._get_leaves(x)

        assert len(train_leaf_indices.shape) == 2, "Leaf indices must be 2D"

    def partition(self, x: np.ndarray) -> np.ndarray:
        leaf_indices = self._get_leaves(x)
        # Leaf indices are the partitions
        return leaf_indices.astype(int)

    @abstractmethod
    def _get_leaves(self, x: np.ndarray) -> np.ndarray:
        pass

    @abstractmethod
    def _fit_model(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: np.ndarray,
        y_val: np.ndarray,
    ):
        pass

    @abstractmethod
    def _prefit_model(self, model: Any, x: np.ndarray) -> None:
        pass


class SKLTreePartition(TreePartition):
    def _truncate_classification(self, leaf_indices: np.ndarray) -> np.ndarray:
        if len(leaf_indices.shape) == 2:
            return leaf_indices

        return leaf_indices[:, :, 0]

    def _get_leaves(self, x: np.ndarray) -> np.ndarray:
        # SKLearn uses the convention that leaf indices start at 1
        leaf_indices = self._truncate_classification(self.model.apply(x))

        # actually rename the leaf indices
        for i in range(leaf_indices.shape[0]):
            for j in range(leaf_indices.shape[1]):
                leaf_indices[i, j] = self.renamer[j][leaf_indices[i, j]]

        return leaf_indices

    def get_leaf_paths(self, x: np.ndarray) -> list[np.ndarray]:
        decision_paths = []

        for tree in self.model.estimators_[:, 0]:
            decision_paths.append(tree.decision_path(x))

        return decision_paths

    def _fit_model(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: np.ndarray,
        y_val: np.ndarray,
    ):
        match self.loss_type:
            case LossType.regression:
                model_class = self.get_regression_model()
            case LossType.binary_classification | LossType.multinomial_classification:
                model_class = self.get_classification_model()
            case _:
                raise ValueError("Loss type not supported")

        self.model = model_class(  # type: ignore[reportOptionalCall]
            n_estimators=self.n_partitions,
            max_leaf_nodes=self.n_cells,
            random_state=self.state,
            **self.model_kwargs,
        )

        self.model.fit(x, y.flatten())

        self._init_leaf_renamer(x)

    def _init_leaf_renamer(self, x: np.ndarray) -> None:
        # Need to use this really hacky method because leaf indices are not
        # sequential

        # Get leaf indices for each data point in each tree
        # SKLearn uses the convention that leaf indices start at 1
        leaf_indices = self._truncate_classification(self.model.apply(x))

        self.renamer = []

        for tree_leaves in leaf_indices.T:
            d = {v: i for i, v in enumerate(np.unique(tree_leaves))}
            self.renamer.append(d)

    def _prefit_model(self, model: Any, x: np.ndarray) -> None:
        self.model = model

        self._init_leaf_renamer(x)

    @abstractmethod
    def get_regression_model(self):
        pass

    @abstractmethod
    def get_classification_model(self):
        pass


class ExtraTreePartition(SKLTreePartition):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        n_cells: int,
        n_partitions: int,
        loss_type: LossType,
        random_state: int = 0,
        prefit_model: Any | None = None,
        x_val: np.ndarray | None = None,
        y_val: np.ndarray | None = None,
        **model_kwargs: dict[str, Any],
    ) -> None:
        super().__init__(
            x,
            y,
            n_cells,
            n_partitions,
            loss_type,
            random_state,
            prefit_model,
            x_val,
            y_val,
            {**model_kwargs, "max_features": 1},
        )

    def get_regression_model(self):
        return ExtraTreesRegressor

    def get_classification_model(self):
        return ExtraTreesClassifier


class LGBMPartition(TreePartition):
    def _get_leaves(self, x: np.ndarray) -> np.ndarray:
        return self.model.predict(x, pred_leaf=True)

    def _fit_model(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: np.ndarray,
        y_val: np.ndarray,
    ):
        match self.loss_type:
            case LossType.regression:
                model_class = LGBMRegressor
            case LossType.binary_classification | LossType.multinomial_classification:
                model_class = LGBMClassifier
            case _:
                raise ValueError("Loss type not supported")

        self.model = model_class(
            n_estimators=self.n_partitions,
            num_leaves=self.n_cells,
            random_state=self.state,
            verbose=-1,
        )

        self.model.fit(x, y.flatten(), eval_set=(x_val, y_val.flatten()))

    def _prefit_model(self, model: Any, x: np.ndarray) -> None:
        self.model = model

    def get_leaf_paths(self, x: np.ndarray) -> list[np.ndarray]:
        raise NotImplementedError("Leaf paths not implemented for LightGBM")


class CBPartition(TreePartition):
    def __init__(
        self,
        x: np.ndarray,
        y: np.ndarray,
        n_cells: int,
        n_partitions: int,
        loss_type: LossType,
        random_state: int = 0,
        prefit_model: Any | None = None,
        x_val: np.ndarray | None = None,
        y_val: np.ndarray | None = None,
        **model_kwargs: dict[str, Any],
    ) -> None:
        super().__init__(
            x,
            y,
            n_cells,
            n_partitions,
            loss_type,
            random_state,
            prefit_model,
            x_val,
            y_val,
            keep_int=True,
            **model_kwargs,
        )

    def _get_leaves(self, x: np.ndarray) -> np.ndarray:
        return self.model.calc_leaf_indexes(x)

    def _fit_model(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: np.ndarray,
        y_val: np.ndarray,
    ):
        match self.loss_type:
            case LossType.regression:
                model_class = CatBoostRegressor
            case LossType.binary_classification | LossType.multinomial_classification:
                model_class = CatBoostClassifier
            case _:
                raise ValueError("Loss type not supported")

        train_pool = Pool(
            x,
            y.flatten(),
            cat_features=self.model_kwargs.get("cat_features"),
        )

        val_pool = Pool(x_val, y_val.flatten())

        max_depth = np.log2(self.n_cells)
        assert max_depth % 1 == 0, "n_cells must be a power of 2"

        self.model = model_class(
            max_depth=int(max_depth),
            iterations=self.n_partitions,
            random_state=self.state,
            **self.model_kwargs,
        )

        self.model.fit(
            train_pool,
            eval_set=val_pool,
            verbose=False,
        )

        self.n_partitions = self.model.tree_count_

    def _prefit_model(self, model: Any, x: np.ndarray) -> None:
        self.model = model

    def get_leaf_paths(self, x: np.ndarray) -> list[np.ndarray]:
        raise NotImplementedError("Leaf paths not implemented for CatBoost")


class RFPartition(SKLTreePartition):
    def get_regression_model(self):
        return RandomForestRegressor

    def get_classification_model(self):
        return RandomForestClassifier

class GBPartition(SKLTreePartition):
    def get_regression_model(self):
        return GradientBoostingRegressor

    def get_classification_model(self):
        return GradientBoostingClassifier

class LinearForestPartition(SKLTreePartition):
    def _fit_model(
        self,
        x: np.ndarray,
        y: np.ndarray,
        x_val: np.ndarray,
        y_val: np.ndarray,
    ):
        assert (np.log2(self.n_cells) % 1) == 0, "n_cells must be a power of 2"

        if "base_estimator" not in self.model_kwargs:
            self.model_kwargs["base_estimator"] = LinearRegression()

        super()._fit_model(x, y, x_val, y_val)

    def get_regression_model(self):
        return LinearForestRegressor

    def get_classification_model(self):
        return LinearForestClassifier



"""
Source: https://gist.github.com/daien/1272551/edd95a6154106f8e28209a1c7964623ef8397246

Proximal operator for l_infinity norm: https://math.stackexchange.com/questions/527872/the-proximal-operator-of-the-l-infty-infinity-norm
"""

import jax.numpy as jnp

FLOAT = jnp.float32
INT = jnp.int32


def euclidean_proj_simplex1(v: jnp.ndarray, s: float = 1.0) -> jnp.ndarray:
    assert s > 0, "Radius s must be strictly positive (%d <= 0)" % s
    shape = v.shape
    # get the array of cumulative sums of a sorted (decreasing) copy of v
    u = -jnp.sort(-v, axis=-1)
    cssv = jnp.cumsum(u, axis=-1, dtype=FLOAT)
    # get the number of > 0 components of the optimal solution
    rho = jnp.sum(
        u * jnp.arange(1, shape[-1] + 1, dtype=FLOAT)[None, :] > (cssv - s),
        axis=-1,
        dtype=INT,
    )
    # compute the Lagrange multiplier associated to the simplex constraint
    theta = jnp.divide(
        (cssv[jnp.arange(0, shape[0], dtype=INT), rho] - s), (rho + 1), dtype=FLOAT
    )
    # compute the projection by thresholding v using theta
    return (v - theta[:, None]).clip(min=0)


def euclidean_proj_l1ball(v: jnp.ndarray, s: float = 1.0) -> jnp.ndarray:
    assert s > 0, "Radius s must be strictly positive (%d <= 0)" % s
    # compute the vector of absolute values
    u = jnp.abs(v)
    # check if v is already a solution
    if u.sum() <= s:
        # L1-norm is <= s
        return v
    # v is not already a solution: optimum lies on the boundary (norm == s)
    # project *u* on the simplex
    w = euclidean_proj_simplex1(u, s=s)
    # compute the solution to the original problem on v
    w *= jnp.sign(v)
    return w



from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from typing import Callable, Sequence

import jax.numpy as jnp
from jax import jit
import matplotlib.pyplot as plt
import numpy as np
from matplotlib import pyplot as plt
from numpy.random import RandomState
from sklearn.exceptions import NotFittedError
from sklearn.metrics import mean_squared_error
from sklearn.preprocessing import StandardScaler
from tqdm import tqdm

from .enums import DTYPE, LossType, NormType
from .partitioner import Partitioner, VPartition, numpify
from .proj_l1_ball import euclidean_proj_l1ball
from .st_loss import LogisticLoss, SquaredError


def core_l21_prox(
    v: jnp.ndarray,
    lam: float,
) -> jnp.ndarray:
    """Compute the proximal operator of the l2,1 norm.

    Args:
        v (jnp.ndarray): The input vector.
        lam (float): The regularization parameter.

    Returns:
        jnp.ndarray: The proximal operator of the l2,1 norm.
    """
    norm = jnp.linalg.norm(v, axis=(0, 2), keepdims=True, ord="fro")
    return jnp.maximum(1 - lam / norm, 0) * v


def task_l21_prox(
    v: jnp.ndarray,
    lam: float,
) -> jnp.ndarray:
    """Compute the proximal operator of the l2,1 norm.

    Args:
        v (jnp.ndarray): The input vector.
        lam (float): The regularization parameter.

    Returns:
        jnp.ndarray: The proximal operator of the l2,1 norm.
    """
    norm = jnp.linalg.norm(v, axis=2, keepdims=True, ord=2)
    return jnp.maximum(1 - lam / norm, 0) * v


def core_l21_norm(
    v: jnp.ndarray,
) -> float:
    """Compute the l2,1 norm where the 2 norm is computed over all cells and tasks.

    Args:
        v (jnp.ndarray): The input vector.
        lam (float): The regularization parameter.

    Returns:
        float: The l2,1 norm.
    """
    return jnp.linalg.norm(v, axis=(0, 2), ord="fro").reshape(-1)


def task_l21_norm(
    v: jnp.ndarray,
) -> jnp.ndarray:
    """Compute the l2,1 norm where the 2 norm is computed over all cells and tasks.

    Args:
        v (jnp.ndarray): The input vector.
        lam (float): The regularization parameter.

    Returns:
        float: The l2,1 norm.
    """
    return jnp.linalg.norm(v, axis=-1, ord=2)


def sq_fnorm(v: jnp.ndarray) -> float:
    """Compute the squared frobenius norm of the input vector.

    Args:
        v (jnp.ndarray): The input vector.

    Returns:
        float: The squared frobenius norm.
    """
    return float(jnp.sum(v**2))


def task_linf1_prox(
    v: jnp.ndarray,
    lam: float,
) -> jnp.ndarray:
    """Compute the linf,1 proximal operator, where the inf norm is computed over
    all cells and each task invidivually.

    Args:
        v (jnp.ndarray): The input vector. Has shape (n_t, n_f, n_c*n_p)
        lam (float): The regularization parameter.

    Returns:
        jnp.ndarray: The projected vector.
    """

    for t in range(v.shape[0]):
        v = v.at[t].set(euclidean_proj_l1ball(v[t] / lam))

    return v


def core_linf1_prox(
    v: jnp.ndarray,
    lam: float,
) -> jnp.ndarray:
    """Compute the linf,1 proximal operator, where the inf norm is computed over
    all cells and tasks together.

    Args:
        v (jnp.ndarray): The input vector.
        lam (float): The regularization parameter.

    Returns:
        jnp.ndarray: The projected vector.
    """
    # (n_t, n_f, n_c*n_p) -> (n_f, n_t * n_c * n_p) and back so that
    # the 1 norm is computed over all cells and tasks. This is easier than trying
    # to compute the proximal operator over a 3D tensor.
    v = jnp.moveaxis(v, 0, 1)
    v_shape = v.shape
    v = v.reshape(v_shape[0], -1)

    v -= lam * euclidean_proj_l1ball(v / lam)

    v = v.reshape(*v_shape)
    return jnp.moveaxis(v, 1, 0)


def core_linf1_norm(
    v: jnp.ndarray,
) -> jnp.ndarray:
    """Compute the linf,1 norm where the inf norm is computed over all cells and
    tasks together.

    Args:
        v (jnp.ndarray): The input vector.

    Returns:
        float: The linf,1 norm.
    """
    return jnp.sum(jnp.max(jnp.abs(v), axis=2), axis=0)


def task_linf1_norm(
    v: jnp.ndarray,
) -> float:
    """Compute the linf,1 norm where the inf norm is computed over all cells and
    tasks together.

    Args:
        v (jnp.ndarray): The input vector.

    Returns:
        float: The linf,1 norm.
    """
    return jnp.max(jnp.abs(v), axis=-1)


class MTJOPLEn:
    def __init__(
        self: MTJOPLEn,
        partitioner: type[Partitioner],
        n_cells: int = 1,
        n_partitions: int = 1,
        random_state: int | RandomState = 0,
        part_kwargs: dict[str, int] | None = None,
        is_regression: bool = True,
    ) -> None:
        if not issubclass(partitioner, Partitioner):
            raise ValueError("Value of partitioner must be a subclass of Partitioner.")

        if n_cells == 1 and n_partitions > 1:
            raise RuntimeWarning(
                "Multiple partitions with a single cell is redundant and will only increase execution time."
            )

        if n_cells < 1:
            raise ValueError("Number of cells must be greater than 0.")
        if n_partitions < 1:
            raise ValueError("Number of partitions must be greater than 0.")

        self.pclass = partitioner
        self.n_cells = n_cells
        self.n_partitions = n_partitions
        self.partitioners: list[Partitioner] | None = None
        self.cws: list[jnp.ndarray] | None = None
        self.cwb: list[jnp.ndarray] | None = None
        self.x_scalers: list[StandardScaler] | None = None
        self.y_scalers: list[StandardScaler] | None = None
        self.part_kwargs: dict[str, int] | None = part_kwargs or {}
        self.is_regression = is_regression
        self.loss = SquaredError() if is_regression else LogisticLoss()

        if not isinstance(random_state, RandomState):
            self.random_state: RandomState = RandomState(random_state)
        else:
            self.random_state: RandomState = random_state

    def _create_partitioners(
        self: MTJOPLEn,
        lst_x: list[np.ndarray],
        lst_y: list[np.ndarray],
    ) -> list[Partitioner]:
        """Create the partitioners for each round.

        Args:
            self (JOPLEn): The JOPLEn object.
            lst_x (list[np.ndarray]): The input data as a list. Each element of
                the list is a numpy array of shape (n_samples, n_features). They
                do not have to have the same dimensionality
            lst_y (list[np.ndarray]): The input labels. See x for more details.
        """
        partitioners = []

        for x, y in tqdm(zip(lst_x, lst_y), total=len(lst_x), desc="Partition tasks"):
            partitioners.append(
                self.pclass(
                    x,
                    y,
                    self.n_cells,
                    self.n_partitions,
                    LossType.regression,
                    self.random_state.randint(0, 2**32 - 1),
                    **self.part_kwargs,
                )
            )

        return partitioners

    def _get_cells(
        self: MTJOPLEn,
        x: np.ndarray,
        task_idx: int,
    ) -> jnp.ndarray:
        """Get the partitions for each round.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (jnp.ndarray): The input data.
            task_idx (int): The index of the task.

        Returns:
            jnp.ndarray: A n_points x n_partitions array of partition indices.
        """
        if self.partitioners is None:
            raise ValueError("Must fit the model before getting partitions.")

        partitioner = self.partitioners[task_idx]
        p_idx = partitioner.partition(x)
        p_idx = p_idx + np.arange(self.n_partitions) * self.n_cells

        binary_mask = np.zeros((x.shape[0], self.n_partitions * self.n_cells))
        binary_mask[np.arange(x.shape[0])[:, None], p_idx] = 1

        return jnp.asarray(binary_mask, dtype=DTYPE)

    def add_bias(self: MTJOPLEn, x: np.ndarray) -> np.ndarray:
        """Add a bias term to the input data.

        Args:
            x (np.ndarray): The input data.

        Returns:
            np.ndarray: The input data with a bias term.
        """
        return np.hstack((x, np.ones((x.shape[0], 1), dtype=DTYPE)))

    def fit(
        self: MTJOPLEn,
        lst_x: list[np.ndarray],
        lst_y: list[np.ndarray],
        lam_core: float = 0.001,
        lam_task: float = 0.001,
        mu: float = 0.001,
        max_iters: int = 1000,
        print_epochs: int = 100,
        verbose: bool = True,
        lst_val_x: list[np.ndarray] | None = None,
        lst_val_y: list[np.ndarray] | None = None,
        threshold: float = 1e-3,
        norm_type: NormType = NormType.L21,
        core_alpha: float = 0.001,
        task_alpha: float = 0.001,
        rel_lr: Sequence[float] | None = None,
    ) -> dict[str, list[float]]:
        """Fit the JOPLEn model.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (list[np.ndarray]): The input data as a list. Each element of the
                list is a numpy array of shape (n_samples, n_features). They do
                not have to have the same dimensionality
            y (list[np.ndarray]): The input labels. See x for more details.
            lam_core (float, optional): The regularization parameter for the
                core weights. Defaults to 0.001.
            lam_task (float, optional): The regularization parameter for the
                task weights. Defaults to 0.001.
            mu (float, optional): The step size. Defaults to 0.001.
            max_iters (int, optional): The maximum number of iterations.
                Defaults to 1000.
            print_epochs (int, optional): The number of epochs between each
                logging. Defaults to 100.
            verbose (bool, optional): Whether to print the logs. Defaults to
                True.
            lst_val_x (Union[list[np.ndarray], None], optional): The validation
                input data. Defaults to None.
            lst_val_y (Union[list[np.ndarray], None], optional): The validation
                input labels. Defaults to None.
            prox (Callable[[jnp.ndarray, float], jnp.ndarray], optional): The
            threshold (float, optional): The threshold for a feature to be
                considered selected. Defaults to 1e-3.
            core_prox (Callable[[jnp.ndarray, float], jnp.ndarray], optional):
                The proximal operator for the core weights. Defaults to
                core_l21_prox.
            task_prox (Callable[[jnp.ndarray, float], jnp.ndarray], optional):
                The proximal operator for the task weights. Defaults to
                task_l21_prox.
        """
        # Rescale/preprocess the data
        lst_x = [numpify(x) for x in lst_x]
        lst_y = [numpify(y) for y in lst_y]

        self.x_scalers = [StandardScaler().fit(x) for x in lst_x]
        self.y_scalers = [StandardScaler().fit(y) for y in lst_y]

        lst_x = [xs.transform(x) for xs, x in zip(self.x_scalers, lst_x)]
        lst_y = [ys.transform(y) for ys, y in zip(self.y_scalers, lst_y)]

        # fit the partitioners and cache the partitions
        self.partitioners = self._create_partitioners(lst_x, lst_y)
        lst_s = [self._get_cells(x, i) for i, x in enumerate(lst_x)]

        lst_x_aug = list(map(self.add_bias, lst_x))

        # Move the training data to the GPU
        lst_x_aug = [jnp.asarray(x, dtype=DTYPE) for x in lst_x_aug]
        lst_y = [jnp.asarray(y, dtype=DTYPE) for y in lst_y]

        del lst_x

        if lst_val_x is not None and lst_val_y is not None:
            lst_val_x = [numpify(x) for x in lst_val_x]
            lst_val_y = [numpify(y) for y in lst_val_y]

            lst_val_x = [xs.transform(x) for xs, x in zip(self.x_scalers, lst_val_x)]
            lst_val_y = [ys.transform(y) for ys, y in zip(self.y_scalers, lst_val_y)]

            lst_val_s = [self._get_cells(x, i) for i, x in enumerate(lst_val_x)]

            lst_val_x = [jnp.asarray(x, dtype=DTYPE) for x in lst_val_x]
            lst_val_y = [jnp.asarray(y, dtype=DTYPE) for y in lst_val_y]

            lst_val_x = list(map(self.add_bias, lst_val_x))
        else:
            lst_val_x = None
            lst_val_y = None
            lst_val_s = None

        lst_wb_prev = jnp.zeros(
            (
                len(lst_x_aug),
                lst_x_aug[0].shape[1],
                self.n_partitions * self.n_cells,
            ),
            dtype=DTYPE,
        )
        lst_ws_prev = lst_wb_prev.copy()

        lst_wb_next = lst_wb_prev
        lst_ws_next = lst_ws_prev

        history = defaultdict(list)

        # Reweight the learning rates for each task
        if rel_lr is None:
            sqrt_ds_sizes = [np.sqrt(x.shape[0]) for x in lst_x_aug]
            rel_lr = jnp.asarray(sqrt_ds_sizes, dtype=DTYPE)
        else:
            assert len(rel_lr) == len(
                lst_x_aug
            ), "Must have a learning rate for each task."
            assert all(lr > 0 for lr in rel_lr), "Learning rates must be positive."

            rel_lr = jnp.asarray(rel_lr, dtype=DTYPE)

        rel_lr /= rel_lr.max()

        # set up the proximal operators and norms
        if norm_type == NormType.L21:
            core_prox = core_l21_prox
            task_prox = task_l21_prox
            core_norm = core_l21_norm
            task_norm = task_l21_norm
        elif norm_type == NormType.LINF1:
            core_prox = core_linf1_prox
            task_prox = task_linf1_prox
            core_norm = core_linf1_norm
            task_norm = task_linf1_norm

        t_curr = 1
        t_next = 1

        # proximal gradient descent
        for i in range(max_iters):
            lst_wb_tmp = lst_wb_next
            lst_ws_tmp = lst_ws_next

            t_next = (1 + jnp.sqrt(1 + 4 * t_curr**2)) / 2
            beta = (t_curr - 1) / t_next

            for j in range(len(lst_x_aug)):
                # Perform accelerated gradient descent
                momentum_b = lst_wb_next[j] + beta * (lst_wb_next[j] - lst_wb_prev[j])
                momentum_s = lst_ws_next[j] + beta * (lst_ws_next[j] - lst_ws_prev[j])

                grad_update = mu * self.loss.grad(
                    momentum_b + momentum_s,
                    lst_x_aug[j],
                    lst_y[j],
                    lst_s[j],
                )

                # lst_wb_next[j] -= rel_lr[j] * grad_update
                # lst_ws_next[j] -= rel_lr[j] * grad_update
                lst_wb_next = lst_wb_next.at[j].set(
                    lst_wb_next[j] - rel_lr[j] * grad_update
                )
                lst_ws_next = lst_ws_next.at[j].set(
                    lst_ws_next[j] - rel_lr[j] * grad_update
                )

                if core_alpha > 0:
                    # lst_wb_next[j, :-1] -= mu * core_alpha * momentum_b[:-1]
                    lst_wb_next = lst_wb_next.at[j].set(
                        lst_wb_next[j] - mu * core_alpha * momentum_b
                    )
                if task_alpha > 0:
                    # lst_ws_next[j, :-1] -= mu * task_alpha * momentum_b[:-1]
                    lst_wb_next = lst_wb_next.at[j].set(
                        lst_wb_next[j] - mu * task_alpha * momentum_s
                    )

            # apply proximal operator
            if lam_core > 0:
                # lst_wb_next[:, :-1] = core_prox(lst_wb_next[:, :-1], mu * lam_core)
                lst_wb_next = lst_wb_next.at[:, :-1].set(
                    core_prox(lst_wb_next[:, :-1], mu * lam_core)
                )
            if lam_task > 0:
                # lst_ws_next[:, :-1] = task_prox(lst_ws_next[:, :-1], mu * lam_task)
                lst_ws_next = lst_ws_next.at[:, :-1].set(
                    task_prox(lst_ws_next[:, :-1], mu * lam_task)
                )

            lst_wb_prev = lst_wb_tmp
            lst_ws_prev = lst_ws_tmp

            # logging
            if (i + 1) % print_epochs == 0:
                res = self.record_performance(
                    lst_wb_next,
                    lst_ws_next,
                    lst_x_aug,
                    lst_y,
                    lst_s,
                    i,
                    verbose,
                    lst_val_x,
                    lst_val_y,
                    lst_val_s,
                    threshold,
                    core_norm,
                    task_norm,
                    lam_core,
                    lam_task,
                    core_alpha,
                    task_alpha,
                )

                for k, v in res.items():
                    if v is not None:
                        history[k].append(v)

        self.cwb = lst_wb_next
        self.cws = lst_ws_next

        return dict(history)

    def record_performance(
        self: MTJOPLEn,
        lst_wb_next: jnp.ndarray,
        lst_ws_next: jnp.ndarray,
        lst_x_aug: list[jnp.ndarray],
        lst_y: list[jnp.ndarray],
        lst_s: list[jnp.ndarray],
        i: int,
        verbose: bool,
        lst_val_x: list[jnp.ndarray] | None,
        lst_val_y: list[jnp.ndarray] | None,
        lst_val_s: list[jnp.ndarray] | None,
        threshold: float,
        core_norm: Callable[[jnp.ndarray], float],
        task_norm: Callable[[jnp.ndarray], float],
        lam_core: float,
        lam_task: float,
        core_alpha: float,
        task_alpha: float,
    ) -> dict[str, tuple[float, ...] | float]:
        tmp_hist = []
        loss_strs = []
        raw_loss = []

        for j in range(len(lst_x_aug)):
            loss_next = self._score(
                lst_wb_next[j] + lst_ws_next[j],
                lst_x_aug[j],
                lst_y[j],
                lst_s[j],
                j,
            )
            raw_loss.append(
                self.loss(
                    lst_wb_next[j] + lst_ws_next[j],
                    lst_x_aug[j],
                    lst_y[j],
                    lst_s[j],
                )
            )

            tmp_hist.append(loss_next)
            loss_strs.append(f"{loss_next:.6f}")

        current_time = datetime.now().strftime("%H:%M:%S")
        report_str = f"[{current_time}]:"

        report_str += f" Epoch {i + 1:>6d} | TrL: {', '.join(loss_strs)}"

        if lst_val_x is not None and lst_val_y is not None:
            tmp_hist_val = []
            val_loss_strs = []
            for j in range(len(lst_val_x)):
                val_loss = self._score(
                    lst_wb_next[j] + lst_ws_next[j],
                    lst_val_x[j],
                    lst_val_y[j],
                    lst_val_s[j],
                    j,
                )
                val_loss_strs.append(f"{val_loss:.6f}")
                tmp_hist_val.append(val_loss)
            report_str += f" | VaL: {', '.join(val_loss_strs)}"
            val_loss = tuple(tmp_hist_val)
        else:
            val_loss = None

        wb_norm = core_norm(lst_wb_next[:, :-1])
        ws_norm = task_norm(lst_ws_next[:, :-1])
        wb_norm_sum = float(wb_norm.sum())
        ws_norm_sum = float(ws_norm.sum())

        report_str += f" | CNorm: {float(wb_norm_sum):.6f}"
        report_str += f" | TNorm: {float(ws_norm_sum):.6f}"

        wb_fnorm = sq_fnorm(lst_wb_next[:, :-1])
        wb_fnorm = sq_fnorm(lst_ws_next[:, :-1])
        report_str += f" | CFNorm: {wb_fnorm:.6f}"
        report_str += f" | TFNorm: {wb_fnorm:.6f}"

        wb_sel_idx = wb_norm > threshold
        ws_sel_idx = ws_norm > threshold
        ws_sel_idx = ws_sel_idx & ~wb_sel_idx
        ws_sel_idx_str = ", ".join([f"{w.sum():>4d}" for w in ws_sel_idx])
        report_str += f" | WbNz: {wb_sel_idx.sum():>4d}"
        report_str += f" | WsNz: {ws_sel_idx_str}"

        b_n_features = float(wb_sel_idx.sum())
        s_n_features = tuple([float(w.sum()) for w in ws_sel_idx])
        train_loss = tuple(tmp_hist)

        # Compensate since all terms are scaled by P for the gradient
        objective = sum(raw_loss)
        objective += lam_core * wb_norm_sum
        objective += lam_task * ws_norm_sum
        objective += core_alpha * wb_fnorm
        objective += task_alpha * wb_fnorm

        if verbose:
            print(report_str)

        return {
            "train_loss": train_loss,
            "raw_loss": raw_loss,
            "val_loss": val_loss,
            "wb_norm": wb_norm_sum,
            "ws_norm": ws_norm_sum,
            "b_n_features": b_n_features,
            "s_n_features": s_n_features,
            "wb_fnorm": wb_fnorm,
            "wb_fnorm": wb_fnorm,
            "objective": objective,
        }

    def _score(
        self: MTJOPLEn,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y_true: jnp.ndarray,
        s: jnp.ndarray,
        task_idx: int,
    ) -> float:
        """Compute the loss function.

        Args:
            self (JOPLEn): The JOPLEn object.
            w (jnp.ndarray): The weight matrix.
            x (jnp.ndarray): The input data.
            y (jnp.ndarray): The input labels.
            s (jnp.ndarray): The partition matrix.
            task_idx (int): The index of the task.

        Returns:
            float: The loss value.
        """
        y_pred = np.array(self.loss.predict(w, x, s))
        y_pred = self.y_scalers[task_idx].inverse_transform(y_pred)

        y_true = self.y_scalers[task_idx].inverse_transform(np.array(y_true))

        return float(mean_squared_error(y_true, y_pred, squared=False))

    def predict(self: MTJOPLEn, x: jnp.ndarray, task_idx: int) -> np.ndarray:
        """Predict the output for the given input and task.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (jnp.ndarray): The input data.
            task_idx (int): The index of the task.

        Returns:
            jnp.ndarray: The predicted output.
        """
        if self.cws is None:
            raise NotFittedError("Must fit the model before predicting.")

        x = numpify(x)

        x = self.x_scalers[task_idx].transform(x)
        s = self._get_cells(x, task_idx)

        x_aug = self.add_bias(x)

        x_aug = jnp.asarray(x_aug)

        y_pred = np.array(
            self.loss.predict(
                self.cwb[task_idx] + self.cws[task_idx],
                x_aug,
                s,
            )
        )
        return self.y_scalers[task_idx].inverse_transform(y_pred)


if __name__ == "__main__":
    import time
    from pathlib import Path
    from pprint import pprint
    from sklearn.datasets import make_regression

    import pandas as pd
    from lightgbm import LGBMRegressor

    # Generate synthetic data for multitask regression
    n_samples = 1000
    n_features = 20
    n_tasks = 3
    noise = 0.1

    np.random.seed(0)

    # Generate synthetic data
    x = np.random.randn(n_samples, n_features)
    tmat = np.random.randn(n_features, n_tasks)
    cmat = np.random.randn(n_features, 1)
    cmat = np.tile(cmat, (1, n_tasks))

    tmat[np.abs(tmat) < 1.5] = 0
    cmat[np.abs(cmat) < 1.5] = 0

    fig, axs = plt.subplots(1, 2, figsize=(10, 5))
    axs[0].matshow(cmat)
    axs[1].matshow(tmat)
    plt.show()

    assert cmat.shape == tmat.shape

    w = cmat + tmat  # Generate weight matrix

    noise_vector = noise * np.random.randn(n_samples, n_tasks)  # Generate noise vector

    # Calculate target variable y using matrix multiplication
    y = np.dot(x, w) + noise_vector

    x_train, x_val, x_test = x[:800], x[800:900], x[900:]
    y_train, y_val, y_test = y[:800], y[800:900], y[900:]

    print(x_train.shape, y_train.shape)

    # Tile the input data for multiple tasks
    x_train = np.tile(x_train, (n_tasks, 1, 1))
    x_val = np.tile(x_val, (n_tasks, 1, 1))
    x_test = np.tile(x_test, (n_tasks, 1, 1))

    # Transpose the targets
    y_train = np.transpose(y_train)
    y_val = np.transpose(y_val)
    y_test = np.transpose(y_test)

    n_cells = 2
    n_partitions = 10
    print_epochs = 100
    lam_task = 0.2
    lam_core = 0.2
    mu = 1e-2
    max_iters = 1000
    norm_type = NormType.L21
    core_alpha = 0.0
    task_alpha = 0.0
    rel_lr = [1] * n_tasks

    def rmse(y_true, y_pred):  # noqa: ANN001, ANN201
        return mean_squared_error(y_true, y_pred, squared=False)

    # get current file path
    path = Path().absolute()

    dummy_pred = []
    for _, ytr, _, yte in zip(x_train, y_train, x_test, y_test):
        dummy = np.mean(ytr)
        y_pred = np.full(yte.shape, dummy)
        dummy_pred.append(rmse(yte, y_pred.flatten()))

    print("Dummy")
    print(dummy_pred)

    lgbm_pred = []
    for xtr, ytr, xte, yte in tqdm(
        zip(x_train, y_train, x_test, y_test), total=n_tasks
    ):
        lgbm = LGBMRegressor(verbose=-1, max_depth=3, n_estimators=100)
        lgbm.fit(xtr, ytr.flatten())
        y_pred = lgbm.predict(xte)
        lgbm_pred.append(rmse(yte, y_pred.flatten()))

    print("LGBM")
    print(lgbm_pred)

    jp = MTJOPLEn(
        VPartition,
        n_cells=n_cells,
        n_partitions=n_partitions,
    )

    start_time = time.time()

    history = jp.fit(
        x_train,
        y_train,
        print_epochs=print_epochs,
        lam_core=lam_core,
        lam_task=lam_task,
        mu=mu,
        max_iters=max_iters,
        verbose=True,
        lst_val_x=x_test,
        lst_val_y=y_test,
        norm_type=norm_type,
        core_alpha=core_alpha,
        task_alpha=task_alpha,
        rel_lr=rel_lr,
    )

    end_time = time.time()

    print("Time:", (end_time - start_time))

    print(len(history["b_n_features"]))

    fig, axs = plt.subplots(1, 3, figsize=(15, 5))
    axs[0].plot(history["raw_loss"])
    axs[1].plot(history["ws_norm"])
    axs[2].plot(history["objective"])

    # set titles
    axs[0].set_title("Raw Training Loss")
    axs[1].set_title("Core Norm")
    axs[2].set_title("Objective Function")

    plt.show()

    core_weights = np.array(jp.cwb)[:, :-1]
    task_weights = np.array(jp.cws)[:, :-1]

    n_tasks, n_feats, n_cells = core_weights.shape
    core_weights = core_weights.transpose(1, 0, 2).reshape(n_feats, -1)
    task_weights = task_weights.transpose(1, 0, 2).reshape(n_feats, -1)

    fig, axs = plt.subplots(1, 2, figsize=(10, 5))
    axs[0].matshow(core_weights)
    axs[1].matshow(task_weights)

    axs[0].set_title("Core Weights")
    axs[1].set_title("Task Weights")

    plt.show()

    # get the selected features for each task
    wb_norm = np.linalg.norm(np.array(jp.cwb), axis=(0, 2), ord="fro")[:-1]
    ws_norm = np.linalg.norm(np.array(jp.cws), axis=2, ord=2)[:, :-1]

    wb_sel_idx = wb_norm > 1e-3
    ws_sel_idx = ws_norm > 1e-3
    ws_sel_idx = ws_sel_idx & ~wb_sel_idx

    x_train = [x[:, wb_sel_idx + ws_sel_idx[i]] for i, x in enumerate(x_train)]
    x_test = [x[:, wb_sel_idx + ws_sel_idx[i]] for i, x in enumerate(x_test)]

    masked_pred = []
    for xtr, ytr, xte, yte in zip(x_train, y_train, x_test, y_test):
        lgbm = LGBMRegressor(verbose=-1)
        lgbm.fit(xtr, ytr.flatten())
        y_pred = lgbm.predict(xte)
        masked_pred.append(rmse(yte, y_pred.flatten()))

    # combine into a table using pandas
    print(
        pd.DataFrame(
            {
                "Dummy": dummy_pred,
                "LGBM": lgbm_pred,
                "LGBM via JOPLEn": masked_pred,
            },
        )
    )



from __future__ import annotations

from abc import ABC, abstractmethod
from enum import Enum, auto
from typing import Any

import jax.numpy as jnp
import numpy as np

from .enums import DTYPE
from .partitioner import Partitioner, TreePartition
from .proj_l1_ball import euclidean_proj_l1ball


class Penalty(ABC):
    def __init__(
        self: Penalty,
        is_smooth: bool,
    ) -> None:
        self.is_smooth = is_smooth

    def build(
        self: Penalty,
        x_train: np.ndarray,
        x_val: np.ndarray,
        partitioner: Partitioner,
    ) -> None:
        pass

    @abstractmethod
    def __call__(
        self: Penalty,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> float:
        pass

    def grad_update(
        self: Penalty,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> jnp.ndarray:
        raise NotImplementedError(
            "Gradient is not implemented, probably because the penalty is not smooth."
        )

    def prox(
        self: Penalty,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> jnp.ndarray:
        raise NotImplementedError(
            "Proximal operator is not implemented, probably because the penalty is smooth."
        )


class SquaredFNorm(Penalty):
    def __init__(
        self: SquaredFNorm,
        lam: float = 1.0,
    ) -> None:
        super().__init__(is_smooth=True)

        assert lam > 0, "The regularization parameter must be positive."
        self.lam = lam

    def __call__(
        self: SquaredFNorm,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> float:
        return self.lam * jnp.sum(w[:bias_idx] ** 2) / 2

    def grad_update(
        self: SquaredFNorm,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> jnp.ndarray:
        w = w.at[:bias_idx].set(w[:bias_idx] - self.lam * mu * w[:bias_idx])
        return w


class NonsmoothGroupPenalty(Penalty, ABC):
    def __init__(
        self: NonsmoothGroupPenalty,
        lam: float = 1.0,
    ) -> None:
        super().__init__(is_smooth=False)

        assert lam > 0, "The regularization parameter must be positive."
        self.lam = lam


class Group21Norm(NonsmoothGroupPenalty):

    def __call__(
        self: Group21Norm,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> float:
        return jnp.linalg.norm(w[:bias_idx], axis=-1, ord=2).sum()

    def prox(
        self: Group21Norm,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> jnp.ndarray:
        norm = jnp.linalg.norm(w[:bias_idx], axis=1, keepdims=True, ord=2)
        w = w.at[:bias_idx].set(w[:bias_idx] * jnp.maximum(1 - self.lam * mu / norm, 0))
        return w


class GroupInf1Norm(NonsmoothGroupPenalty):

    def __call__(
        self: GroupInf1Norm,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> float:
        return jnp.sum(jnp.max(jnp.abs(w[:bias_idx]), axis=-1))

    def prox(
        self: GroupInf1Norm,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> jnp.ndarray:
        w = w.at[:bias_idx].set(
            -self.lam * mu * euclidean_proj_l1ball(w[:bias_idx] / self.lam * mu)
        )

        return w


class NuclearNorm(NonsmoothGroupPenalty):

    def __call__(
        self: NuclearNorm,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> float:
        s = jnp.linalg.svd(w[:bias_idx], full_matrices=True, compute_uv=False)
        return jnp.sum(s)

    def prox(
        self: NuclearNorm,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> jnp.ndarray:
        u, s, v = jnp.linalg.svd(w[:bias_idx], full_matrices=False)
        s = jnp.maximum(s - self.lam * mu, 0)
        w = w.at[:bias_idx].set(u @ (s[:, None] * v))
        return w


class L1Norm(Penalty):

    def __init__(
        self: L1Norm,
        lam: float = 1.0,
    ) -> None:
        super().__init__(is_smooth=False)

        assert lam > 0, "The regularization parameter must be positive."
        self.lam = lam

    def __call__(
        self: L1Norm,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> float:
        return self.lam * jnp.sum(jnp.abs(w[:bias_idx]))

    def prox(
        self: L1Norm,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        *_: list,
    ) -> jnp.ndarray:
        s = jnp.sign(w[:bias_idx])
        thresh = jnp.maximum(jnp.abs(w[:bias_idx]) - self.lam * mu, 0)
        # w[:bias_idx] = s * thresh
        w = w.at[:bias_idx].set(s * thresh)
        return w


class LaplacianType(Enum):
    STANDARD = auto()
    LEFT_NORMALIZED = auto()
    NORMALIZED = auto()


class DistanceWeight(ABC):
    def __init__(
        self: DistanceWeight,
        x_train: np.ndarray,
        x_val: np.ndarray,
        partitioner: Partitioner,
        **params: dict[str, Any],
    ) -> None:
        self.train_weights = self.weight(x_train, partitioner, **params)

        if x_val is not None:
            self.val_weights = self.weight(x_val, partitioner, **params)
        else:
            self.val_weights = None

    def __call__(
        self: DistanceWeight,
        is_train: bool,
    ) -> jnp.ndarray:
        return self.train_weights if is_train else self.val_weights

    @abstractmethod
    def weight(
        self: DistanceWeight,
        x: np.ndarray,
        partitioner: Partitioner,
        **params: dict,
    ) -> np.ndarray:
        raise NotImplementedError("The weight function is not implemented.")


class RBFWeight(DistanceWeight, ABC):
    def __init__(
        self: RBFWeight,
        x_train: np.ndarray,
        x_val: np.ndarray,
        partitioner: Partitioner,
        sigma: float = 1.0,
    ) -> None:
        assert sigma > 0, "Sigma must be positive."

        self.sigma = sigma

        super().__init__(x_train, x_val, partitioner, sigma=sigma)


class EuclidRBFWeight(RBFWeight, ABC):
    def distance(
        self: EuclidRBFWeight,
        x: np.ndarray,
        partitioner: Partitioner,
    ) -> np.ndarray:
        return np.linalg.norm(x[:, None] - x[None, :], axis=-1) ** 2


class TreeRBFWeight(RBFWeight, ABC):
    def __init__(
        self: RBFWeight,
        x_train: jnp.ndarray,
        x_val: jnp.ndarray,
        partitioner: Partitioner,
        sigma: float = 1,
    ) -> None:
        assert issubclass(
            type(partitioner), TreePartition
        ), "The partitioner must be a tree partitioner."

        super().__init__(x_train, x_val, partitioner, sigma)

    def distance(
        self: TreeRBFWeight,
        x: np.ndarray,
        partitioner: Partitioner,
    ) -> np.ndarray:
        leaf_paths = partitioner.get_leaf_paths(x)

        distances = np.empty((len(leaf_paths), x.shape[0], x.shape[0]))

        for i, path in enumerate(leaf_paths):
            path_sum = path.sum(axis=1).A1  # converts to a dense 1D numpy array
            path_dot_path_T = path.dot(path.T).toarray()

            distance = path_sum[:, None] + path_sum[None, :] - 2 * path_dot_path_T
            # distances[i] = distance
            distances = distances.at[i].set(distance)

        return distances.mean(axis=0)


class EuclidGaussWeight(EuclidRBFWeight):
    def weight(
        self: EuclidGaussWeight,
        x: np.ndarray,
        partitioner: Partitioner,
        sigma: float,
    ) -> np.ndarray:
        return np.exp(-self.distance(x, partitioner) * sigma)


class TreeGaussWeight(TreeRBFWeight):
    def weight(
        self: TreeGaussWeight,
        x: np.ndarray,
        partitioner: Partitioner,
        sigma: float,
    ) -> np.ndarray:
        return np.exp(-self.distance(x, partitioner) * sigma)


class EuclidMultiQuadWeight(EuclidRBFWeight):
    def weight(
        self: EuclidMultiQuadWeight,
        x: np.ndarray,
        partitioner: Partitioner,
        sigma: float,
    ) -> np.ndarray:
        weights = self.distance(x, partitioner)
        mask = weights != 0
        weights = np.divide(1, 1 + weights * sigma, where=mask)
        weights = weights.at[~mask].set(0)
        return weights


class TreeMultiQuadWeight(TreeRBFWeight):
    def weight(
        self: TreeMultiQuadWeight,
        x: np.ndarray,
        partitioner: Partitioner,
        sigma: float,
    ) -> np.ndarray:
        weights = self.distance(x, partitioner)
        mask = weights != 0
        weights = np.divide(1, 1 + weights * sigma, where=mask)
        weights = weights.at[~mask].set(0)
        return weights


class Laplacian(Penalty, ABC):
    def __init__(
        self: Laplacian,
        is_smooth: bool,
        lam: float = 1.0,
        sigma: float = 1.0,
        weight_class: type[DistanceWeight] = EuclidGaussWeight,
        laplacian_type: LaplacianType = LaplacianType.STANDARD,
    ) -> None:
        super().__init__(is_smooth=is_smooth)

        assert lam > 0, "The regularization parameter must be positive."
        assert sigma > 0, "The variance parameter must be positive."

        self.lam = lam
        self.sigma = sigma
        self.laplacian_type = laplacian_type
        self.weight_class = weight_class

    def create_laplacian(self: Laplacian, weights: np.ndarray) -> jnp.ndarray:
        if self.laplacian_type == LaplacianType.LEFT_NORMALIZED:
            s = np.sum(weights, axis=1)
            d_inv = np.reciprocal(s, where=s != 0)
            L = d_inv[:, None] * weights
        elif self.laplacian_type == LaplacianType.NORMALIZED:
            s = np.sqrt(np.sum(weights, axis=1))
            d_inv = np.reciprocal(s, where=s != 0)
            L = d_inv[:, None] * weights * d_inv[None, :]
        elif self.laplacian_type == LaplacianType.STANDARD:
            L = weights
        else:
            raise ValueError(f"Invalid Laplacian type: {self.laplacian_type}")

        return jnp.array(L, dtype=DTYPE)

    def build(
        self: Laplacian,
        x_train: np.ndarray,
        x_val: np.ndarray,
        partitioner: Partitioner,
    ) -> None:
        self.weight = self.weight_class(x_train, x_val, partitioner, self.sigma)

        self.L_train = self.create_laplacian(self.weight(True))
        self.L_val = self.create_laplacian(self.weight(False))


class SquaredLaplacian(Laplacian):
    def __init__(
        self: SquaredLaplacian,
        **kwargs: dict,
    ) -> None:
        super().__init__(
            is_smooth=True,
            **kwargs,
        )

    def __call__(
        self: SquaredLaplacian,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> float:
        L = self.L_train if is_train else self.L_val

        return self.lam * float(y_pred.T @ L @ y_pred / (2 * x.shape[0]))

    def grad_update(
        self: SquaredLaplacian,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> jnp.ndarray:
        L = self.L_train if is_train else self.L_val

        # L may not be symmetric
        res = x.T @ ((L @ y_pred) * s)
        res += x.T @ ((L.T @ y_pred) * s)

        return w - self.lam * mu * res / (2 * x.shape[0])


class TVLaplacian(Laplacian):
    def __init__(
        self: TVLaplacian,
        **kwargs: dict,
    ) -> None:
        assert (
            "laplacian_type" not in kwargs
        ), "The Laplacian type is not applicable to the total variation penalty."

        super().__init__(
            is_smooth=False,
            laplacian_type=LaplacianType.STANDARD,
            **kwargs,
        )

    def __call__(
        self: TVLaplacian,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> float:
        L = self.L_train if is_train else self.L_val

        return (
            w - self.lam * jnp.mean(jnp.abs(y_pred[:, None] - y_pred[None, :]) * L) / 2
        )

    def prox(
        self: TVLaplacian,
        mu: float,
        w: jnp.ndarray,
        bias_idx: int | None,
        x: jnp.ndarray,
        y_pred: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> jnp.ndarray:
        raise NotImplementedError("The total variation penalty is not implemented.")



from enum import Enum, auto
import jax.numpy as jnp

# import cupy

# DTYPE = cupy.float32
DTYPE = jnp.float32


class LossType(Enum):
    multinomial_classification = auto()
    binary_classification = auto()
    regression = auto()
    ranking = auto()


class CellModel(Enum):
    constant = auto()
    linear = auto()


class NormType(Enum):
    """Norm types that are currently supported."""

    L21 = auto()
    LINF1 = auto()



from __future__ import annotations

from collections import defaultdict
from datetime import datetime
from typing import Any

import numpy as np
from numpy.random import RandomState
from sklearn.exceptions import NotFittedError
from sklearn.preprocessing import StandardScaler

from .enums import CellModel, LossType, DTYPE
from .partitioner import Partitioner, numpify
from .st_loss import Loss, SquaredError
from .st_penalty import (
    Penalty,
)
import jax.numpy as jnp


class JOPLEn:

    def __init__(
        self: JOPLEn,
        partitioner: type[Partitioner] | Partitioner,
        n_cells: int = 1,
        n_partitions: int = 1,
        random_state: int | RandomState = 0,
        loss_fn: type[Loss] = SquaredError,
        cell_model: CellModel = CellModel.linear,
        x_scaler: StandardScaler | None = None,
        y_scaler: StandardScaler | None = None,
        regularizers: list[Penalty] | None = None,
        part_kwargs: dict[str, float] | None = None,
        mu: float = 0.001,
        max_iters: int = 1000,
        patience: int = 100,
        stop_thresh: float = 1e-6,
        early_stop: bool = True,
        rescale: bool = True,
        mu_decr: float = 0.5,
        use_nesterov: bool = True,
    ) -> None:
        if isinstance(partitioner, type):
            cond = issubclass(partitioner, Partitioner)
        else:
            cond = isinstance(partitioner, Partitioner)

        if not cond:
            raise ValueError("Value of partitioner must be a subclass of Partitioner.")

        if n_cells == 1 and n_partitions > 1:
            raise RuntimeWarning(
                "Multiple partitions with a single cell is redundant and will only increase execution time."
            )

        if n_cells < 1:
            raise ValueError("Number of cells must be greater than 0.")
        if n_partitions < 1:
            raise ValueError("Number of partitions must be greater than 0.")

        if isinstance(partitioner, type):
            self.pclass = partitioner
            self.partitioner: Partitioner | None = None
        else:
            self.pclass = type(partitioner)
            self.partitioner = partitioner

        assert (x_scaler is None) == (
            y_scaler is None
        ), "Either both scalers must be None or neither should be."
        self.x_scaler: StandardScaler | None = x_scaler
        self.y_scaler: StandardScaler | None = y_scaler

        self.n_cells = n_cells
        self.n_partitions = n_partitions
        self.w: jnp.ndarray | None = None
        self.loss_fn: LossType = loss_fn()
        self.bl_regr: bool = self.loss_fn.loss_type == LossType.regression
        self.cell_model: CellModel = cell_model
        self.regularizers = regularizers if regularizers else []

        self.smooth_reg = [r for r in self.regularizers if r.is_smooth]
        self.nonsmooth_reg = [r for r in self.regularizers if not r.is_smooth]
        self.mu = mu
        self.max_iters = max_iters
        self.patience = patience
        self.stop_thresh = stop_thresh
        self.early_stop = early_stop
        self.mu_decr = mu_decr
        self.use_nesterov = use_nesterov
        self.rescale = rescale

        self.part_kwargs: dict[str, float] | None = part_kwargs if part_kwargs else {}

        if not isinstance(random_state, RandomState):
            self.random_state: RandomState = RandomState(random_state)
        else:
            self.random_state: RandomState = random_state

    def _create_partitioner(
        self: JOPLEn,
        x: np.ndarray,
        y: np.ndarray,
        val_x: np.ndarray | None,
        val_y: np.ndarray | None,
    ) -> Partitioner:
        """Create the partitioner for each round.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (np.ndarray): The input data, a numpy array of shape (n_samples,
            n_features).
            y (np.ndarray): The input labels. See x for more details.
        """
        partitioner = self.pclass(
            x,
            y,
            self.n_cells,
            self.n_partitions,
            self.loss_fn.loss_type,
            int(self.random_state.randint(0, 2**32 - 1)),
            x_val=val_x,
            y_val=val_y,
            **self.part_kwargs,
        )

        self.n_partitions = partitioner.n_partitions

        return partitioner

    def _get_cells(
        self: JOPLEn,
        x: np.ndarray,
    ) -> jnp.ndarray:
        """Get the partitions for each round.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (jnp.ndarray): The input data.

        Returns:
            jnp.ndarray: A n_points x n_partitions array of partition indices.
        """
        if self.partitioner is None:
            raise ValueError("Must fit the model before getting partitions.")

        p_idx = self.partitioner.partition(x)

        p_idx = p_idx + np.arange(self.n_partitions) * self.n_cells

        binary_mask = np.zeros((x.shape[0], self.n_partitions * self.n_cells))
        print(binary_mask.shape, x.shape, p_idx.shape)
        binary_mask[np.arange(x.shape[0])[:, None], p_idx] = 1

        return jnp.asarray(binary_mask, dtype=DTYPE)

    def add_bias(self: JOPLEn, x: np.ndarray) -> tuple[np.ndarray, int | None]:
        """Add a bias term to the input data.

        Args:
            x (np.ndarray): The input data.

        Returns:
            tuple[np.ndarray, int|None]: The input data with the bias term added,
            and an index that can be used to remove the bias term.
        """

        match self.cell_model:
            case CellModel.constant:
                return np.ones((x.shape[0], 1), dtype=DTYPE), None
            case CellModel.linear:
                return np.hstack((x, np.ones((x.shape[0], 1), dtype=DTYPE))), -1
            case _:
                raise ValueError("Invalid cell model.")

    def _score(
        self: JOPLEn,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y_true: jnp.ndarray,
        s: jnp.ndarray,
    ) -> float:
        """Compute the loss function.

        Args:
            self (JOPLEn): The JOPLEn object.
            w (jnp.ndarray): The weight matrix.
            x (jnp.ndarray): The input data.
            y (jnp.ndarray): The input labels.
            s (jnp.ndarray): The partition matrix.

        Returns:
            float: The loss value.
        """
        y_pred = np.array(self.loss_fn.predict(w, x, s))

        if self.bl_regr and self.y_scaler is not None:
            y_pred = self.y_scaler.inverse_transform(y_pred)
            y_true = self.y_scaler.inverse_transform(np.array(y_true))
        else:
            y_true = np.array(y_true)

        return self.loss_fn.error(y_true, y_pred)

    def predict(self: JOPLEn, x: jnp.ndarray) -> np.ndarray:
        """Predict the output for the given input and task.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (jnp.ndarray): The input data.

        Returns:
            jnp.ndarray: The predicted output.
        """
        if self.w is None:
            raise NotFittedError("Must fit the model before predicting.")

        x = numpify(x)

        if self.x_scaler is not None:
            x = self.x_scaler.transform(x)
        s = self._get_cells(x)

        x, _ = self.add_bias(x)

        x = jnp.asarray(x, dtype=DTYPE)

        y_pred = np.array(self.loss_fn.predict(self.w, x, s))

        if self.bl_regr and self.y_scaler is not None:
            return self.y_scaler.inverse_transform(y_pred)
        else:
            return y_pred

    def compute_obj_terms(
        self: JOPLEn,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
        is_train: bool,
    ) -> tuple[dict[str, float], float]:
        objective_terms = {}

        y_pred = self.loss_fn.predict(w, x, s)

        objective_terms["loss"] = self.loss_fn(w, x, y, s)

        for reg in self.regularizers:
            name = reg.__class__.__name__
            objective_terms[name] = float(reg(w, self.bias_idx, x, y_pred, s, is_train))

        return objective_terms, sum(objective_terms.values())

    def compute_obj_grad(
        self: JOPLEn,
        mu: float,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        if len(self.smooth_reg) == 0:
            return w

        y_pred = self.loss_fn.predict(w, x, s)

        w_out = w.copy()

        # apply each smooth regularizer as a gradient update
        for reg in self.smooth_reg:
            w_out = reg.grad_update(mu, w, self.bias_idx, x, y_pred, s, True)

        return w_out

    def compute_obj_prox(
        self: JOPLEn,
        mu: float,
        w: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
    ) -> jnp.ndarray:
        if len(self.nonsmooth_reg) == 0:
            return w

        y_pred = self.loss_fn.predict(w, x, s)

        # Use the proximal average to apply multple proximal operators
        tmp_weights = jnp.zeros_like(w, dtype=DTYPE)

        for reg in self.nonsmooth_reg:
            tmp_weights += reg.prox(mu, w, self.bias_idx, x, y, y_pred, s, True)

        return tmp_weights / len(self.nonsmooth_reg)

    def fit(
        self: JOPLEn,
        x: np.ndarray,
        y: np.ndarray,
        print_epochs: int = 100,
        verbose: bool = True,
        val_x: np.ndarray | None = None,
        val_y: np.ndarray | None = None,
    ) -> dict[str, list[float]]:
        """Fit the JOPLEn model.

        Args:
            self (JOPLEn): The JOPLEn object.
            x (np.ndarray): The input data, a numpy array of shape (n_samples,
                n_features).
            y (np.ndarray): The input labels. See x for more details.
            core (float, optional): The regularization parameter for the regularization.
            mu (float, optional): The step size. Defaults to 0.001.
            max_iters (int, optional): The maximum number of iterations.
                Defaults to 1000.
            print_epochs (int, optional): The number of epochs between each
                logging. Defaults to 100.
            verbose (bool, optional): Whether to print the logs. Defaults to
                True.
            val_x (Union[np.ndarray, None], optional): The validation
                input data. Defaults to None.
            val_y (Union[np.ndarray, None], optional): The validation
                input labels. Defaults to None.
            feat_thresh (float, optional): The feat_thresh for a feature to be
                considered selected. Defaults to 1e-3.
            core_prox (Callable[[jnp.ndarray, float], jnp.ndarray], optional):
                The proximal operator for the weights. Defaults to l21_prox.
        """
        # Rescale/preprocess the data
        x = numpify(x)
        y = numpify(y)

        mu = self.mu

        if self.rescale:
            if self.x_scaler is None:
                self.x_scaler = StandardScaler().fit(x)
                if self.bl_regr:
                    self.y_scaler = StandardScaler().fit(y)

            x = self.x_scaler.transform(x)
            if self.bl_regr:
                y = self.y_scaler.transform(y)

        # fit the partitioner and cache the partitions
        # This check allows us to set the partitioner externally for experiments
        if self.partitioner is None:
            self.partitioner = self._create_partitioner(x, y, val_x, val_y)

        for reg in self.regularizers:
            reg.build(x, val_x, self.partitioner)

        s = self._get_cells(x)

        x, self.bias_idx = self.add_bias(x)

        # Move the training data to the GPU
        x = jnp.asarray(x, dtype=DTYPE)
        y = jnp.asarray(y, dtype=DTYPE)

        using_val = val_x is not None and val_y is not None
        if using_val:
            val_x = numpify(val_x)
            val_y = numpify(val_y)

            if self.rescale:
                val_x = self.x_scaler.transform(val_x)
                if self.bl_regr:
                    val_y = self.y_scaler.transform(val_y)

            val_s = self._get_cells(val_x)

            val_x, _ = self.add_bias(val_x)

            val_x = jnp.asarray(val_x, dtype=DTYPE)
            val_y = jnp.asarray(val_y, dtype=DTYPE)
        else:
            val_x = None
            val_y = None
            val_s = None

        w_prev = jnp.zeros(
            (
                x.shape[1],
                self.n_partitions * self.n_cells,
            ),
            dtype=DTYPE,
        )
        w_next = w_prev
        m_next = w_prev

        history = defaultdict(lambda: defaultdict(list))

        best_val = np.inf
        best_val_w = w_prev
        best_val_idx = 0
        obj_prev = self.loss_fn(w_next, x, y, s)

        best_test_w = w_prev

        t_curr = 1
        t_next = 1
        # proximal gradient descent
        for i in range(self.max_iters):
            w_tmp = w_next
            # Perform accelerated gradient descent
            if self.use_nesterov:
                t_next = (1 + jnp.sqrt(1 + 4 * t_curr**2)) / 2
                beta = (t_curr - 1) / t_next
                m_next = w_next + beta * (w_next - w_prev)

                # loss gradient
                w_next = m_next - mu * self.loss_fn.grad(m_next, x, y, s)
            else:
                w_next = w_prev - mu * self.loss_fn.grad(w_prev, x, y, s)

            # apply each smooth regularizer as a gradient update
            w_next = self.compute_obj_grad(mu, w_next, x, y, s)
            w_next = self.compute_obj_prox(mu, w_next, x, y, s)

            # Compute the objective function
            obj_terms, obj_next = self.compute_obj_terms(w_next, x, y, s, True)

            w_prev = w_tmp

            # logging
            if (i + 1) % print_epochs == 0:
                res = self.record_performance(
                    w_next,
                    x,
                    y,
                    s,
                    i,
                    verbose,
                    val_x,
                    val_y,
                    val_s,
                )

                for k, v in res.items():
                    for kk, vv in v.items():
                        history[k][kk].append(vv)

            if self.early_stop and using_val and i > 0:
                _, val_obj = self.compute_obj_terms(w_next, val_x, val_y, val_s, False)

                if best_val > val_obj:
                    best_val = val_obj
                    best_val_w = w_next
                    best_val_idx = i
                elif (i - best_val_idx > self.patience) and (
                    best_val - val_obj < self.stop_thresh
                ):
                    break
            else:
                best_val_w = w_next
                best_val_idx = i

            if obj_next > obj_prev:
                # if t_next == 1 and t_next == 1:
                mu *= self.mu_decr

                w_next = best_test_w
                w_prev = best_test_w
                t_next = 1
                t_curr = 1
            else:
                best_test_w = w_next
                t_curr = t_next

            if jnp.abs(obj_next - obj_prev) / jnp.abs(obj_prev) <= self.stop_thresh:
                break

        self.w = best_val_w

        return dict({k: dict(v) for k, v in history.items()})

    def record_performance(
        self: JOPLEn,
        w_next: jnp.ndarray,
        x: jnp.ndarray,
        y: jnp.ndarray,
        s: jnp.ndarray,
        i: int,
        verbose: bool,
        val_x: jnp.ndarray | None,
        val_y: jnp.ndarray | None,
        val_s: jnp.ndarray | None,
    ) -> Any:
        current_time = datetime.now().strftime("%H:%M:%S")
        report_str = f"[{current_time}]:"

        train_loss = self._score(w_next, x, y, s)
        report_str += f" Epoch {i + 1:>6d} | TrL: {train_loss:.6f}"

        if val_x is not None and val_y is not None:
            val_loss = self._score(
                w_next,
                val_x,
                val_y,
                val_s,
            )
            report_str += f" | VaL: {val_loss:.6f}"
        else:
            val_loss = None

        output = {
            "train": {
                "loss": train_loss,
            },
            "val": {
                "loss": val_loss,
            },
        }

        train_obj_terms, train_obj = self.compute_obj_terms(w_next, x, y, s, True)

        output["train"].update(train_obj_terms)

        if val_x is not None and val_y is not None:
            val_obj_terms, val_obj = self.compute_obj_terms(
                w_next, val_x, val_y, val_s, False
            )
            output["val"].update(val_obj_terms)

        # Compensate since all terms are scaled by P for the gradient
        report_str += f" | Obj: {train_obj:.6f}"

        if verbose:
            print(report_str)

        return output

    def get_weights(self: JOPLEn) -> np.ndarray:
        """Get the weights of the model.

        Returns:
            np.ndarray: The weights of the model.
        """
        if self.w is None:
            raise NotFittedError("Must fit the model before getting weights.")

        return self.w.get()


if __name__ == "__main__":
    import matplotlib.pyplot as plt
    from sklearn.datasets import make_regression
    from sklearn.metrics import mean_squared_error
    from sklearn.model_selection import train_test_split
    from JOPLEn.partitioner import RFPartition

    from JOPLEn.st_penalty import (
        Group21Norm,
        GroupInf1Norm,
        L1Norm,
        SquaredFNorm,
        SquaredLaplacian,
    )

    # Generate regression data
    x, y = make_regression(
        n_samples=1000,
        n_features=20,
        n_informative=10,
        n_targets=1,
        noise=0.0,
        random_state=0,
    )

    # Split the dataset into training and test sets
    x_train, x_test, y_train, y_test = train_test_split(
        x, y, test_size=0.2, random_state=0
    )

    print(x.shape)

    # Further split the test set into validation and test sets
    x_val, x_test, y_val, y_test = train_test_split(
        x_test,
        y_test,
        test_size=0.5,
        random_state=0,  # Adjust the test_size to get 25% of the original for validation
    )

    jp = JOPLEn(
        partitioner=RFPartition,
        n_cells=20,
        n_partitions=100,
        loss_fn=SquaredError,
        cell_model=CellModel.constant,
        mu=1e-3,
        max_iters=1000,
        early_stop=True,
        rescale=False,
        regularizers=[
            SquaredFNorm(lam=0.1),
            # L1Norm(lam=0.01),
            # SquaredLaplacian(lam=0.1, laplacian_type=LaplacianType.LEFT_NORMALIZED),
            # Group21Norm(lam=0.01),
        ],
    )

    history = jp.fit(
        x_train,
        y_train,
        val_x=x_val,
        val_y=y_val,
        print_epochs=10,
    )

    y_pred = jp.predict(x_test)

    print("JOPLEn MSE:", mean_squared_error(y_test, y_pred, squared=False))

    print(
        "Naive MSE:",
        mean_squared_error(
            y_test, y_train.mean() * np.ones_like(y_test), squared=False
        ),
    )

    n_terms = len(history["train"].keys())
    fig, axs = plt.subplots(1, n_terms, figsize=(5 * n_terms, 5))

    if n_terms == 1:
        axs = [axs]

    for ax, (k, v) in zip(axs, history["train"].items()):
        ax.plot(v)
        ax.set_title(k)

    plt.show()
